from npiai.core.base import App
from npiai_proto import api_pb2
from typing_extensions import override


class GitHub(App):

    def __init__(self, access_token: str, npi_endpoint: str = None):
        super().__init__(
            app_name="github",
            app_type=api_pb2.GITHUB,
            endpoint=npi_endpoint,
        )
        self.access_token = access_token

    @override
    def authorize(self):
        super()._authorize(credentials={"access_token": self.access_token})



import json
import threading
import traceback
from typing import List
import grpc
import uuid
from abc import ABC, abstractmethod

from openai import OpenAI
from openai.types.chat import (
    ChatCompletionSystemMessageParam,
    ChatCompletionToolChoiceOptionParam,
    ChatCompletionToolParam,
    ChatCompletionUserMessageParam,
)

from npiai_proto import api_pb2_grpc, api_pb2
from npiai.utils import logger
from npiai.core import hitl


class App(ABC):
    __app_name: str
    __app_type: api_pb2.AppType
    __npi_endpoint: str
    stub: api_pb2_grpc.AppServerStub
    hitl_handler: hitl.HITLHandler = None

    def __init__(
            self,
            app_name: str,
            app_type: api_pb2.AppType,
            endpoint: str = "localhost:9140",
            hitl_handler: hitl.HITLHandler = None,
            npi_token: str = None,
            insecure: bool = True,
    ):
        self.__app_name = app_name
        if endpoint is None:
            endpoint = "localhost:9140"
        self.__npi_endpoint = endpoint
        self.__app_type = app_type
        if insecure:
            channel = grpc.insecure_channel(self.__npi_endpoint)
        else:
            channel = grpc.secure_channel(target=self.__npi_endpoint, credentials=grpc.ssl_channel_credentials())
        self.stub = api_pb2_grpc.AppServerStub(channel)
        self.hitl_handler = hitl_handler
        self.__npi_token = npi_token

    def tool_name(self):
        return self.__app_name

    def schema(self):
        try:
            resp = self.stub.GetAppSchema(
                request=api_pb2.AppSchemaRequest(
                    type=self.__app_type
                ),
                metadata=self.__get_metadata(),
            )
        except Exception as e:
            logger.error(e)
            return None
        return json.loads(resp.schema)

    def chat(self, msg: str) -> str:
        resp = self.stub.Chat(
            request=api_pb2.Request(
                code=api_pb2.RequestCode.CHAT,
                chat_request=api_pb2.ChatRequest(
                    type=self.__app_type,
                    instruction=msg
                ),
            ),
            metadata=self.__get_metadata(),
        )
        while True:
            match resp.code:
                case api_pb2.ResponseCode.FINISHED:
                    return resp.chat_response.message
                case api_pb2.ResponseCode.MESSAGE:
                    logger.info(f'[{self.__app_name}]: Received message: {resp.chat_response.message}')
                    resp = self.stub.Chat(
                        request=api_pb2.Request(
                            code=api_pb2.RequestCode.FETCH,
                            request_id=str(uuid.uuid4()),
                            thread_id=resp.thread_id,
                            chat_request=api_pb2.ChatRequest(
                                type=self.__app_type,
                            )
                        ),
                        metadata=self.__get_metadata(),
                    )
                case api_pb2.ResponseCode.SUCCESS:
                    resp = self.stub.Chat(
                        request=api_pb2.Request(
                            code=api_pb2.RequestCode.FETCH,
                            request_id=str(uuid.uuid4()),
                            thread_id=resp.thread_id,
                            chat_request=api_pb2.ChatRequest(
                                type=self.__app_type,
                            )
                        ),
                        metadata=self.__get_metadata(),
                    )
                case api_pb2.ResponseCode.ACTION_REQUIRED:
                    resp = self.stub.Chat(request=self.__call_human(resp), metadata=self.__get_metadata())
                case _:
                    logger.error(f'[{self.__app_name}]: Error: failed to call function, unknown response code {resp.code}')
                    raise Exception("Error: failed to call function")

    def hitl(self, handler: hitl.HITLHandler):
        self.hitl_handler = handler

    def authorize(self):
        pass

    def _authorize(self, credentials: dict[str, str]):
        self.stub.Authorize(
            request=api_pb2.AuthorizeRequest(
                type=self.__app_type,
                credentials=credentials,
            ),
        )
        logger.info(f'[{self.__app_name}]: Authorized')

    def __call_human(self, resp: api_pb2.Response) -> api_pb2.Request:
        human_resp = self.hitl_handler.handle(
            hitl.convert_to_hitl_request(
                req=resp.action_response,
                app_name=self.__app_name
            )
        )
        if human_resp is hitl.ACTION_APPROVED:
            result = "approved"
        else:
            result = "denied"
        return api_pb2.Request(
            code=api_pb2.RequestCode.ACTION_RESULT,
            request_id=str(uuid.uuid4()),
            thread_id=resp.thread_id,
            action_result_request=api_pb2.ActionResultRequest(
                action_id=resp.action_response.action_id,
                action_result=result,
            )
        )

    def __get_metadata(self):
        return (('x-host', 'test.playground.npi.ai'),
                ('x-npi-token', self.__npi_token))


class Agent:
    __agent_name: str
    __npi_endpoint: str
    __prompt: str
    __description: str
    __llm: OpenAI
    tool_choice: ChatCompletionToolChoiceOptionParam = "auto"
    fn_map: dict = {}
    hitl_handler: hitl.HITLHandler = None

    def __init__(
            self,
            agent_name: str,
            description: str,
            prompt: str,
            endpoint: str = None,
            llm: OpenAI = None,
            hitl_handler: hitl.HITLHandler = None,
    ):
        self.__agent_name = agent_name
        self.__description = description
        self.__prompt = prompt
        self.__npi_endpoint = endpoint
        self.__llm = llm
        self.hitl_handler = hitl_handler

    def use(self, *apps: App):
        for app in apps:
            # inherit HITL handler
            if app.hitl_handler is None:
                app.hitl_handler = self.hitl_handler
            app_name = app.tool_name()
            self.fn_map[app_name] = app

    def group(self, *agents: 'Agent'):
        for agent in agents:
            app_name = agent.__as_app().tool_name()
            self.fn_map[app_name] = agent

    def hitl(self, handler: hitl.HITLHandler):
        self.hitl_handler = handler

    def run(self, msg: str) -> str:

        messages = [ChatCompletionSystemMessageParam(
            content=self.__prompt,
            role="system",
        ), ChatCompletionUserMessageParam(
            content=msg,
            role="user",
        )]

        while True:
            response = self.__llm.chat.completions.create(
                model="gpt-4-turbo-preview",
                messages=messages,
                tools=self.__tools(),
                tool_choice=self.tool_choice,
                max_tokens=4096,
            )
            response_message = response.choices[0].message

            messages.append(response_message)

            if response_message.content:
                logger.info(response_message.content + '\n')

            tool_calls = response_message.tool_calls

            if tool_calls is None:
                return response_message.content

            for tool_call in tool_calls:
                fn_name = tool_call.function.name
                args = json.loads(tool_call.function.arguments)

                try:
                    res = self.__call(fn_name, args)
                    # logger.info(f"Response: {res}")
                    logger.debug(f'[{self.__agent_name}]: app `{fn_name}` returned: {res}')
                except Exception as e:
                    res = ''.join(traceback.format_exception(e))
                    logger.error(res)

                messages.append(
                    {
                        "tool_call_id": tool_call.id,
                        "role": "tool",
                        "name": fn_name,
                        "content": res,
                    }
                )

    def when(self, task: str):
        running = True
        msg = f'When {task}'

        def _poll():
            nonlocal running
            while running:
                self.run(msg)

        def _stop():
            nonlocal running
            running = False

        thread = threading.Thread(target=_poll)
        # thread.daemon = True
        thread.start()

        return _stop

    def __call(self, fn_name: str, args: dict) -> str:
        call_msg = f'[{self.__agent_name}]: Calling {fn_name}({args})'
        logger.info(call_msg)

        if fn_name not in self.fn_map:
            raise Exception(f"Tool {fn_name} not found.")

        fn = self.fn_map[fn_name]
        return fn.chat(args['message'])

    def __tools(self) -> List[ChatCompletionToolParam]:
        tools = []
        for tool_name, tool in self.fn_map.items():
            tools.append(tool.schema())
        return tools

    def __as_app(self) -> App:
        return App(self.__agent_name, api_pb2.AppType.CHAT)



from termcolor import colored
from npiai.core.hitl import (
    HITLHandler, HITLRequest, HITLResponse, ActionRequestCode, ActionResultCode, ACTION_APPROVED, ACTION_DENIED)

from npiai_proto import api_pb2


class ConsoleHandler(HITLHandler):
    def handle(self, req: HITLRequest) -> HITLResponse:
        match req.code:
            case ActionRequestCode.INFORMATION:
                print(colored(f"[{req.app_name}]: Additional information Required", 'green', attrs=['bold']))
                print(colored(f"[{req.app_name}]: Agent Request: ", 'green', attrs=['bold']))
                human_response = input(colored(f'[{req.app_name}]: Type Your Response: ', 'magenta'))
                return HITLResponse(
                    code=ActionResultCode.SUCCESS,
                    msg=human_response,
                )
            case ActionRequestCode.CONFIRMATION:
                print(colored(f"[{req.app_name}]: Action request for approving", 'green', attrs=['bold']))
                print(colored(f"[{req.app_name}]: Action Detail:\n\n{req.message}", 'green', attrs=['bold']))
                human_response = input(colored(f'[{req.app_name}]: [Yes/No?] > ', 'magenta'))
                if human_response.lower() == "yes":
                    resp = ACTION_APPROVED
                else:
                    resp = ACTION_DENIED
                human_response = input(colored(f'[{req.app_name}]: Do you want give a message back? (typing Enter to '
                                               f'skip) > ', 'magenta'))
                if human_response is not None:
                    resp.message = human_response
                return resp
        return ACTION_DENIED

    def type(self) -> api_pb2.ActionType:
        return api_pb2.ActionType.CONSOLE



import datetime
import time

from twilio.rest import Client
import loguru

from npiai.core.hitl import (
    HITLHandler, HITLRequest, HITLResponse, ActionRequestCode, ACTION_APPROVED, ACTION_DENIED)

from npiai_proto import api_pb2


class TwilioHandler(HITLHandler):
    def __init__(self, account_id: str, auth_token: str,
                 from_number: str, to_number: str, sms: bool = True):
        self.account_id = account_id
        self.auth_token = auth_token
        self.from_phone_number = from_number
        self.client = Client(account_id, auth_token)
        self.to_number = to_number
        self.sms = sms

    def handle(self, req: HITLRequest) -> HITLResponse:
        match req.code:
            case ActionRequestCode.INFORMATION:
                notice = f'[{req.app_name}]: Additional information Required\n'
                notice += f'[{req.app_name}]: Agent Request: '
                notice += f'{req.message}'
            case ActionRequestCode.CONFIRMATION:
                notice = f'[{req.app_name}]: Action request for approving\n'
                notice += f'[{req.app_name}]: Action Detail:\n\n{req.message} '
                notice += f'[{req.app_name}]: [Yes/No?]'
            case _:
                return ACTION_DENIED

        message = self.client.messages.create(
            from_=self.from_phone_number,
            body=notice,
            to=self.to_number,
        )
        print(f'Message has been successfully sent, sid: {message.sid}')
        human_response = self.__wait_reply(datetime.datetime.utcnow())
        if human_response.lower() == "yes":
            resp = ACTION_APPROVED
        else:
            resp = ACTION_DENIED
        return resp

    def type(self) -> api_pb2.ActionType:
        return api_pb2.ActionType.CONSOLE

    def __wait_reply(self, after: datetime) -> str:
        print('Waiting for a reply...')
        while True:
            msgs = self.client.messages.list(from_=self.to_number, date_sent_after=after)
            if len(msgs) > 0:
                loguru.logger.info(f'received reply from: {msgs[0].from_}, body: {msgs[0].body}')
                return msgs[0].body
            time.sleep(1)



from abc import ABC, abstractmethod
from enum import Enum

from npiai_proto import api_pb2


class ActionRequestCode(Enum):
    UNKNOWN_ACTION = 0
    INFORMATION = 1
    SINGLE_SELECTION = 2
    MULTIPLE_SELECTION = 3
    CONFIRMATION = 4


class ActionResultCode(Enum):
    APPROVED = 0
    DENIED = 1
    MESSAGE = 2
    SUCCESS = 3


class HITLRequest:
    def __init__(self, code: ActionRequestCode, message: str, app_name: str = 'Unknown'):
        self.code = code
        self.app_name = app_name
        self.message = message


class HITLResponse:
    def __init__(self, code: ActionResultCode, msg: str = None):
        self.code = code
        self.message = msg


ACTION_APPROVED = HITLResponse(ActionResultCode.APPROVED)
ACTION_DENIED = HITLResponse(ActionResultCode.DENIED)


class HITLHandler(ABC):
    @abstractmethod
    def handle(self, req: HITLRequest) -> HITLResponse:
        pass

    @abstractmethod
    def type(self) -> api_pb2.ActionType:
        pass


def convert_to_hitl_request(req: api_pb2.ActionRequiredResponse, app_name: str = None) -> HITLRequest:
    match req.type:
        case api_pb2.ActionType.INFORMATION:
            return HITLRequest(
                code=ActionRequestCode.INFORMATION,
                message=req.message,
                app_name=app_name,
            )
        case api_pb2.ActionType.CONFIRMATION:
            return HITLRequest(
                code=ActionRequestCode.CONFIRMATION,
                message=req.message,
                app_name=app_name,
            )
        case api_pb2.ActionType.SINGLE_SELECTION:
            return HITLRequest(
                code=ActionRequestCode.SINGLE_SELECTION,
                message=req.message,
                app_name=app_name,
            )
        case api_pb2.ActionType.MULTIPLE_SELECTION:
            return HITLRequest(
                code=ActionRequestCode.MULTIPLE_SELECTION,
                message=req.message,
                app_name=app_name,
            )
    return HITLRequest(
        code=ActionRequestCode.UNKNOWN_ACTION,
        message="invalid action request",
        app_name=app_name,
    )



from .google import Calendar, Gmail
from .human_feedback import ConsoleFeedback
from .discord import Discord
from .github import GitHub

__all__ = ['Calendar', 'Gmail', 'Discord', 'GitHub', 'ConsoleFeedback']



from typing_extensions import override

from openai import AssistantEventHandler, Client
from npiai.core.toolset import ToolSet


class EventHandler(AssistantEventHandler):
    def __init__(self, toolset: ToolSet, llm: Client, thread_id: str, stream_handler=None):
        super().__init__()
        self.toolset = toolset
        self.llm = llm
        self.thread_id = thread_id
        self.stream_handler = stream_handler

    @override
    def on_event(self, event):
        # Retrieve events that are denoted with 'requires_action'
        # since these will have our tool_calls
        if event.event == 'thread.run.requires_action':
            run_id = event.data.id  # Retrieve the run ID from the event data
            self.handle_requires_action(event.data, run_id)

    def handle_requires_action(self, data, run_id):
        tool_outputs = []

        for tool in data.required_action.submit_tool_outputs.tool_calls:
            tool_outputs.append({
                "tool_call_id": tool.id,
                "output": self.toolset.call(tool),
            })

        # Submit all tool_outputs at the same time
        self.submit_tool_outputs(tool_outputs, run_id)

    def submit_tool_outputs(self, tool_outputs, run_id):
        # Use the submit_tool_outputs_stream helper
        with self.llm.beta.threads.runs.submit_tool_outputs_stream(
                thread_id=self.thread_id,
                run_id=run_id,
                tool_outputs=tool_outputs,
                event_handler=EventHandler(self.toolset, self.llm, self.thread_id),
        ) as stream:
            self.stream_handler(run_id, stream)



from npiai.core.base import App
from npiai.core.hitl import HITLHandler
from openai import Client

from npiai.tools.hitl import EmptyHandler
import json
from typing import List


class ToolSet:

    def __init__(self, tools: List[App], llm: Client, hitl_handler: HITLHandler = None, ):
        self._llm = llm
        self._hitl_handler = hitl_handler
        if self._hitl_handler is None:
            self._hitl_handler = EmptyHandler()

        if len(tools) == 0:
            raise Exception("At least one tool is required.")

        self._tools = {}
        for tool in tools:
            tool.authorize()
            tool.hitl(self._hitl_handler)
            self._tools[tool.tool_name()] = tool

    def openai(self):
        tools = []
        for tool in self._tools.values():
            tools.append(tool.schema())
        return tools

    def call(self, body) -> str:
        tool = self._tools.get(body.function.name)
        if tool is None:
            return "Tool not found"
        params = json.loads(body.function.arguments)

        return tool.chat(params['message'])



from npiai.core.hitl import (
    HITLHandler, HITLRequest, HITLResponse, ActionRequestCode, ActionResultCode, ACTION_APPROVED, ACTION_DENIED)

from npiai_proto import api_pb2


class EmptyHandler(HITLHandler):
    def handle(self, req: HITLRequest) -> HITLResponse:
        match req.code:
            case ActionRequestCode.INFORMATION:
                return HITLResponse(
                    code=ActionResultCode.SUCCESS,
                    msg="No human can response your help",
                )
            case ActionRequestCode.CONFIRMATION:
                return HITLResponse(
                    code=ActionResultCode.APPROVED,
                    msg="automatically approved",
                )
        return ACTION_DENIED

    def type(self) -> api_pb2.ActionType:
        return api_pb2.ActionType.CONSOLE
